<?php

/**
 * PHPOF2/DBRow.class.php - PHPOF2 database row class (Active Record pattern)
 *
 * Copyright 2001-2010 Tim Jackson (tim@timj.co.uk)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU Lesser General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @package PHPOF2
 *
 */

require_once 'PHPOF2/DBTable.php';

/**
 * LARGE OBJECT (BINARY DATA) HANDLING:
 * This is the way large objects (LOBs; typically binary data) works:
 * If the column is EMPTY, NULL or ZERO there is no binary data in the database
 * Any (internally-generated) 'SELECT' query will NOT place the actual binary
 * data in an object; it will replace it with the placeholder constant
 * PHPOF2_DB::LOB_PLACEHOLDER. To retrieve the data, it will be necessary to
 * call PHPOF2_DBRow::getLOBData() with the column name you want to retrieve 
 * data from.
 * 
 * When inserting binary data, you need to FIRST have the row which will contain
 * the data already inserted in the database. (This is for cross-database
 * compatibility where some databases required that the row exists first before
 * you can add LOB data to it). Then you need to call 
 * PHPOF2_DBRow::storeLOBData() which will insert the data directly into the
 * database and fill the object attribute for the relevant column with
 * PHPOF2_DB::LOB_PLACEHOLDER
 * 
 * Ditto for updating binary data in a row.
 * 
 * If you want to remove the binary data from a row, mark the column with
 * PHPOF2_DB::LOB_DELETEPENDING and next time you do a DBRow::update() the data
 * will be deleted. Alternatively, call PHPOF2_DBRow::deleteLOBData()
*/
class PHPOF2_DBRow {	// friend DB_result

	const LOB_GET = 'lob_get';
	const LOB_CHECKONLY = 'lob_check';

	const COMPKEY_SEPARATOR = '#!#'; // The separator used when serializing composite keys

	// for move();
	const MOVE_UP = 'up';
	const MOVE_DOWN = 'down';
	const SORT_TOP = 'top';
	const SORT_BOTTOM = 'bottom';

	public     $db;	            // pointer to DB object. needs to be accessed from driver
	public     $table;          // Contains table object which this row belongs to. needs to be accessed from driver
	
	// Fields to be copied from DBTable
	protected  $fieldlist;                // Comma-separated fieldlist of all fields EXCEPT binary fields or Pri Keys
	protected  $pri_keys = array();       // Array containing names of all primary keys
	protected  $fields_normal = array();  // Array containing names of all columns except LOB fields
	protected  $fields_lob = array();     // array containing names of all LOB fields
	protected  $original_values = null;
	protected  $record_changes = false;
	protected  $driver;
	
	// This is set to TRUE on a sucessful get()/getUnique();
	public     $retrieved_ok = false;
	
	// Used internally to know what row this "originally" came from in case user
	// wants to update the pri key
	private    $_orig_prikey = null;
	
	// Internal use. Fields to skip when doing a clear() - allows optimisation
	// internally so that functions don't have to reselect primary key columns
	// when they already know them
	private    $_clear_skip_fields = array();
	
	/**
	 * Used as a marker in the database when LOB fields are saved to the filesystem
	 */
	const      LOB_FILE_MARKER = '!#FILE#!:';
	
	/**
	* @desc Initialising constructor. $table should be an object of type DBTable
	* @access public
	* @return void
	*/
	public function __construct (PHPOF2_DBTable &$table, $prikey_value = NULL)
	{
		// Store a link to the table object
		$this->table = &$table;
		
		// Store a link to the database
		$this->db = $this->table->db;
		
		// Find out what the DB-specific driver is for this instance
		$dbsyntax = $table->db->getDSN('array');
		$dbsyntax = $dbsyntax['dbsyntax'];
		$this->driver = 'PHPOF2_Driver_DBRow_' . $dbsyntax;
		require_once("PHPOF2/Driver/DBRow/$dbsyntax.php");
		
		// Copy some useful stuff over from the table object
		$this->fields_normal = $this->table->getFieldListNormal();
		$this->fields_lob = $this->table->getFieldListLOB();
		
		$this->pri_keys = $this->table->getPrimaryKeys();
		
		// Generate generally-useful field list
		$fieldlist = array();
		foreach ($this->fields_normal as $field) {
			$fieldlist[] = $this->db->quoteIdentifier($field);
		}
		
		$this->fieldlist = implode(',', $fieldlist);
		
		// Initialise the properties
		$this->clear();
		
		settype($key,'string');	// This is necessary to make sure that numeric keys match below expression
		if ($prikey_value !== NULL) {
			$this->setPrimaryKeyValue($prikey_value);
			$this->get();	
		}
		return true;
	}
	
	
	/**
	 * @desc Set the row's primary key using a serialized value
	 * 
	 * @param string $serialized_value
	 */
	public function setPrimaryKeyValue($serialized_value)
	{
		
		$key_values = $this->unserialisePriKey($serialized_value);
		foreach ($key_values as $col_name => $value)
		{
			$this->$col_name = $value;
		}
	}
	
	
	/**
	 * @desc Return a serialised (string) version of the VALUES stored in the primary key(s) in the table
	 * 
	 * @return string - COMPKEY_SEPARATOR separated list of pri key vals
	 */
	public function serialisePriKey()
	{
		if (count($this->pri_keys) == 0 || !$this->primaryKeyFilled()) {
			return null;
		}
		$keys = array();
		foreach($this->pri_keys as $field){
			$keys[] = $this->{$field};
		}
		return implode(self::COMPKEY_SEPARATOR, $keys);
	}
	
	/**
	 * Get a string which is a serialised version of a key
	 *
	 * @param array $values Array of column-value pairs that you want to serialised, in the form colname => colval. The order is not important.
	 */
	public function getSerialisedPriKey(array $values)
	{
		$keys = array();
		foreach($this->pri_keys as $field) {
			if (isset($values[$field])) {
				$keys[] = $values[$field];
			} else {
				throw new Exception ("Value for primary key column '$field' is missing from the array of values passed to getSerialisedPriKey()");
			}
		}
		return implode(self::COMPKEY_SEPARATOR, $keys);
	}
	
	
	/**
	 * @desc Return an unserialised (assoc array) version of VALUES for the primary key(s)
	 * 
	 * @param string $value  Primary key values, in the correct order and COMPKEY_SEPARATOR separated
	 * @return array         Associative array of key_column => value
	 */
	public function unserialisePriKey($value)
	{
		$values = explode(self::COMPKEY_SEPARATOR, $value);
		foreach ($this->pri_keys as $index => $key_col)
		{
			if (!isset($values[$index])) {
				throw new Exception('unserialise failed - serialised key does not have correct number of key fields');
			}
			$output[$key_col] = $values[$index];
		}
		return $output;
	}
	
	
	/**
	 * Retrieve a row from the database based on a specified primary key. 
	 * To get by a value other than the primary key, use getUnique() instead.
	 *
	 * @param mixed $prikey_value   The primary key value of the row to get.
	 * @return bool                 TRUE if a row was successfully retrieved; FALSE otherwise.
	 */
	public function get($prikey_value = NULL)
	{
		if (count($this->pri_keys) == 0) {
			throw new Exception ('PHPOF2_DBRow::get() called but table has no primary key');
		}
				
		$this->retrieved_ok = false;

		if ($prikey_value === NULL) {
			$prikey_value = $this->serialisePriKey();
		}
		$this->clear();
		$this->setPrimaryKeyValue($prikey_value);
		
		// First deal with non-LOB fields
		$sql = 'SELECT ' . $this->fieldlist . ' FROM '.$this->table->name . ' WHERE ' . $this->getPrikeyWhereClause();
		$res = $this->db->query ($sql);
		
		if ($res->numRows()> 1) {
			throw new Exception ('PHPOF2_DBRow::get() called but primary key value ("' . $prikey_value . '") was not unique in table "' . $this->table->name . '"');
		} else if ($res->numRows() > 0) {
			$this->_clear_skip_fields = $this->pri_keys;
			$this->fillFromResult($res);
			$this->_clear_skip_fields = array();
			$this->retrieved_ok = true;
			$this->initChangeRecording();
			return true;
		}
		return false;
	}
	
	/**
	 * Retrieve a unique row from the database based on an array of 
	 * field/value pairs. Returns TRUE if a unique row was successfully retrieved,
	 * otherwise FALSE.
	 * The public property retrieved_ok is also set similarly.	
	 *
	 * @param  array $field_value_pairs  List of field/value pairs which the unique row to be retrieved should match 
	 * @return bool  Whether or not a unique row was retrieved
	 * @throws Exception
	*/
	public function getUnique (array $field_value_pairs)
	{
		$this->retrieved_ok = false;
		$this->clear();		
		
		//Check the input
		if (count($field_value_pairs) == 0) {
			throw new Exception('Field/value pairs array passed to PHPOF2_DBRow::getUnique() was empty, could not retrieve values');
		}
		
		// Construct the WHERE clause
		$wheres = array();
		foreach ($field_value_pairs as $field_tmp => $value_tmp){
			$wheres[] = $field_tmp.'='.$this->db->quote($value_tmp) . "\n";
		}
		$unique_where = implode(' AND ', $wheres);
	
		// We need to select the primary key values too, if they exist
		$fields_to_select = array_merge($this->pri_keys, $this->fields_normal);
		
		// Select matching rows from the DB
		$sql = 'SELECT '.implode(',', $fields_to_select) .	' FROM ' . $this->table->name . ' WHERE ' . $unique_where;
		
		$res = $this->db->query ($sql);
		
		if ($res->numRows() > 1) {
			// The provided field/value pairs obviously weren't unique
			throw new Exception ('More than one record retrieved by PHPOF2_DBRow::getUnique method');
		} else if ($res->numRows() == 1) {
			$this->_clear_skip_fields = $this->pri_keys;
			$this->fillFromResult($res);
			$this->_clear_skip_fields = array();
			$this->retrieved_ok = true;
			return true;
		}
		
		return false;
	}

	/**
	 * Get a WHERE clause for use in SQL queries based on the primary keys in the table
	 * If serialised key is passed, this is used for constructing the clause. 
	 * Otherwise, the current values of the prikey columns are used
	 *
	 * @param  string $serialised_key Optional serialised primary key
	 * @return string                 WHERE clause
	 */
	public function getPrikeyWhereClause($serialised_key=null)	// This is public because it is needed from the database drivers
	{
		if ($serialised_key !== null) {
			$keys = $this->unserialisePriKey($serialised_key);
			foreach ($this->pri_keys as $prikey) {
				$prikey_where[] = $prikey . '=' . $this->db->quote($keys[$prikey]);
			}
		} else {
			foreach ($this->pri_keys as $prikey) {
				$prikey_where[] = $prikey . '=' . $this->db->quote($this->$prikey);
			}
		}
		return implode(' AND ', $prikey_where);
	}

	/**
	 * Run through the binary fields in the row, see if there is data in them in the DB and mark with placeholders as appropriate
	 * This is intended primarily as an internal function and it should not 
	 * normally be necessary for user code to call it.
	 */
	public function resolveLOBFields()
	{
		return call_user_func_array(array($this->driver, 'resolveLOBFields'), array(&$this));
	}
	
	/**
	 * Clears all object fields which relate to columns in the database. (Sets
	 * them to NULL)
	 * 
	 * @return void
	 */
	public function clear ()
	{
		$fields = array_merge($this->pri_keys, $this->fields_lob, $this->fields_normal);
		if (count($fields) > 0) {
			foreach ($fields as $field) {
				if (!in_array($field, $this->_clear_skip_fields)) {
					$this->$field = null;
				}
			}
		}
		$this->initChangeRecording();
	}
	
	/**
	 * Returns the contents of $column_name directly - ideal for use with LOB
	 * types
	 * 
	 * @param $column_name string The name of the column to fetch data from
	 * @return mixed
	*/
	public function getLOBData ($column_name, $query=PHPOF2_DBRow::LOB_GET)
	{
		$field_list_lob = $this->table->getFieldListLOB();
		if (!in_array($column_name, $field_list_lob)) {
			throw new Exception("Column $column_name is not a LOB field - don't call DBRow::getLOBData() on it");
		}
		$sql = "SELECT $column_name".
				 ' FROM '.$this->table->name.
				 ' WHERE '.$this->getPrikeyWhereClause();
		$res = $this->db->query ($sql);
		if ($res->numRows() > 0) {
			$resultarray = $res->fetchRow(MDB2_FETCHMODE_ASSOC);
			if ($query == self::LOB_CHECKONLY) { 
				if (empty($resultarray[$column_name])) {
					return false;
				} else {
					return true;
				}
			} else {
				$data = call_user_func_array(array($this->driver, 'retrieveLOBData'), array(&$this, $column_name, $resultarray));

				// Check for a file marker
				if (substr($data, 0, strlen(self::LOB_FILE_MARKER)) === self::LOB_FILE_MARKER) {
					$fs_path = $this->table->getLOBStorageFSPath($column_name);
					if ($fs_path === false) {
						throw new Exception("Cannot find filesystem storage location for table " . $this->table->name . ", column '$column_name': check 'lob_storage_fs_path' option");
					}
					$file_name = $fs_path . DIRECTORY_SEPARATOR . substr($data, strlen(self::LOB_FILE_MARKER));

					if (!is_file($file_name)) {
						throw new Exception("File '$file_name' does not exist");
					} elseif(!is_readable($file_name)) {
						throw new Exception("File '$file_name' is not readable");
					}
					$data = file_get_contents($file_name);
				}
				return $data;
			}
		} else {
			return false;
		}
	}
	
	/**
	 * Is primary key filled? Works with both single and composite keys
	 * 
	 * @return bool
	 */
	public function primaryKeyFilled()
	{
		if (count($this->pri_keys) == 0) {
			throw new Exception('PHPOF2_DBRow::primaryKeyFilled() called on object with no primary key');
		}
		
		$filled = true;
		foreach ($this->pri_keys as $key_field)
		{
			if ($this->$key_field === null) $filled = false;
		}
		return $filled;
	}
	
	/**
	 * Stores given binary raw data into a column
	 *
	 * @param $data string The binary data itself
	 * @param $data_column string The name of the column which will receive the binary data from the uploaded file
	 * @param $storage_type string  A PHPOF2_DB::LOB_STORAGE_* constant, to force a particular type of storage
	 * @return bool
	 */
	public function storeLOBData(&$data, $data_column, $storage_type=null)
	{
		if (!$this->primaryKeyFilled()) {
			throw new Exception('DBRow::storeLOBData() called on object which has not yet been stored in the database');
			return false;
		}
		if ($storage_type === null) {
			$storage_type = $this->table->getLOBStorageType($data_column);
		}
		$fs_path = $this->table->getLOBStorageFSPath($data_column);
		if ($storage_type == PHPOF2_DB::LOB_STORAGE_DB) {
			call_user_func_array(array($this->driver, 'storeLOBDataFromBuffer'), array(&$this, $data_column, &$data));
			
			// If the data was previously in the filesystem then remove
			$old_filename = $this->getLOBStorageFSFilename($data_column, $this->_orig_prikey);
			if ($old_filename !== false) {
				$old_filename_full = $fs_path . DIRECTORY_SEPARATOR . $old_filename;
				if (file_exists($old_filename_full)) {
					unlink($old_filename_full);
				}
			}
		} else if ($storage_type == PHPOF2_DB::LOB_STORAGE_FS) {
			if ($fs_path === false) throw new Exception('Could not get LOB storage filesystem path for table ' . $this->table->name);
			if (!is_dir($fs_path) || !is_writable($fs_path)) {
				throw new Exception("LOB storage path $fs_path for table " . $this->table->name . " is not writeable");
			}
			
			$new_filename = $this->getLOBStorageFSFilename($data_column);
			$old_filename = $this->getLOBStorageFSFilename($data_column, $this->_orig_prikey);
					
			// Check if the key has changed; if so remove old file
			if ($old_filename !== false && $old_filename != $new_filename) {
				$old_filename_full = $fs_path . DIRECTORY_SEPARATOR . $old_filename;
				if (file_exists($old_filename_full)) {
					unlink($old_filename_full);
				}
			}
			if ($data !== null) {
				$new_filename_full = $fs_path . DIRECTORY_SEPARATOR . $new_filename;
			
				$out = file_put_contents($new_filename_full, $data);
				if ($out === false) {
					throw new Exception("Error writing LOB data to file $new_filename_full");
				}

				// Store the LOB file marker and relative path to the actual data file in the LOB column
				$db_placeholder = self::LOB_FILE_MARKER . $new_filename;
				call_user_func_array(array($this->driver, 'storeLOBDataFromBuffer'), array(&$this, $data_column, &$db_placeholder));
			} else {
				// Data is null, so there's no point in writing a file
				// It would also lead to behaviour inconsistent with data stored in
				// the DB as subsequent discovery of the row would mean the (null)
				// column would have the LOB_PLACEHOLDER in it.
				// So we just store a NULL into the raw database column
				$null = null;
				call_user_func_array(array($this->driver, 'storeLOBDataFromBuffer'), array(&$this, $data_column, &$null));
			}
		} else {
			throw new Exception("Unknown LOB storage type for column '$data_column': $storage_type");
		}
		
		$this->$data_column = null; // Make it consistent with when data is retrieved initially
		if ($data !== null) {
			$this->$data_column = PHPOF2_DB::LOB_PLACEHOLDER;
		}
		return true;
	}
	
	/**
	 * Returns a filename for the specified LOB data column, if it is stored on the filesystem. If 'prikey' is specified, for the given primary key, otherwise with the current values present in the object
	 *
	 * @param string $data_column Name of the column
	 * @param string $prikey Serialised primary key
	 */
	public function getLOBStorageFSFilename($data_column, $prikey=null)
	{
		if ($prikey === null) {
			return md5($this->table->name . $this->serialisePriKey() . $data_column);
		} else {
			$sql = "SELECT $data_column".
			 ' FROM ' . $this->table->name.
			 ' WHERE ' . $this->getPrikeyWhereClause($prikey) .
			 ' AND SUBSTRING(CONVERT(' . $data_column . ' USING utf8), 1, ' . strlen(self::LOB_FILE_MARKER).') = ' . $this->db->quote(self::LOB_FILE_MARKER);
			$res = $this->db->query ($sql);
			if ($res->numRows() > 0) {
				$resultarray = $res->fetchRow(MDB2_FETCHMODE_ASSOC);
				return substr($resultarray[$data_column], strlen(self::LOB_FILE_MARKER));
			}
			return false;
		}
	}
	
	/**
	 * Fill the properties of current object from an associative array, where
	 * the array contains "column name" => "column content" pairs
	 * 
	 * @param array $array Array of data to be stored
	 * @return void
	*/
	public function fillFromArray (array $array)
	{
		if (count($array) > 0) {
			foreach ($array as $key => $value) {
				$this->$key = $value;
			}
		}
		$this->initChangeRecording();
	}
	
	/**
	 * Store a row in the database. If the row corresponding to the current
	 * primary key already exists, it is updated. Otherwise a new row is 
	 * inserted. For LOB data, you may additionally need to call 
	 * @see PHPOF2_DBRow::storeLOBData()
	 * 
	 * @return void
	 */
	public function store ()
	{
		if (count($this->pri_keys) > 0) {           // If the table has a primary key column
			if ($this->table->key_auto_increment) {  // ...then if the table has an auto-incrementing key column
				if ($this->primaryKeyFilled()) {      // ......then if the key is supplied
					$this->update();                   // .........Update an existing row according to the key
				} else {                              // ......Otherwise, 
					$this->insert();                   // .........Insert a new row (creating a new key id)
				}
			} else {                                 // There is no auto-inc key column,
				if ($this->primaryKeyFilled()) {      // so check if a key is supplied.
					if ($this->rowExists()) {          // ...If it is, see if a row exists with that key.
						$this->update();                // ......If it does, update that row.
					} else {                           // ...Otherwise,
						$this->insert();                // ......Insert a new row with the given key
					}
				} else {                              // There's no auto-inc key column and no key was supplied
					// don't know what to do
					throw new Exception('PHPOF2_DBRow::store() called but no key supplied and key column is not auto-incrementing'); 
				}
			}
		} else {
			throw new Exception('PHPOF2_DBRow::store() called on a table with no primary key column(s)');
		}	
	}
	
	/**
	 * Find out if a row exists in the database with a primary key matching the values currently held in the object
	 *
	 * @return bool
	 */
	public function rowExists ()
	{
		$res = $this->db->query('SELECT 1 FROM ' . $this->table->name . ' WHERE ' . $this->getPrikeyWhereClause());
		if (PEAR::isError($res)) {
    		throw new Exception($res->getMessage());
		}
		if ($res->numRows() >0 ) {
			return true;
		}
		return false;
	}
	
	/**
	 * Insert a new row into the database based on the values currently stored in the properties of the object
	 *
	 * @return void
	 */
	public function insert ()
	{
		// Check to see if user is trying to handle binary fields incorrectly
		foreach ($this->fields_lob as $field)
		{
			if ($this->$field !== null && 
				$this->$field !== PHPOF2_DB::LOB_PLACEHOLDER &&
				$this->$field !== PHPOF2_DB::LOB_DELETEPENDING) {
				
				throw new Exception("Field '$field' contains binary data. Don't insert binary data like this.");
			}
		}
		
		### PHPOF-DYNAMIC-START: DBROW-INSERT-SQL ###
		
		$fields_to_insert = $this->fields_normal;
		// If the table has a key, but it is not auto-incrementing, add the key column name to the list of fields to be inserted
		$pri_keys_count = count($this->pri_keys);
		if ($pri_keys_count == 1) {
			if (!$this->table->key_auto_increment || isset($this->{$this->pri_keys[0]})) {
				$fields_to_insert[] = $this->pri_keys[0];
			} // TODO: else remove key from list of fields??
		} else if ($pri_keys_count > 1) {
			// if more than 1 pri key, we assume no auto-incrementing as this does not really makes sense
			$fields_to_insert = array_merge($fields_to_insert, $this->pri_keys);			
		}
		
		// Add a sort order if this table has a sort field and the current item doesn't have one
		if (!empty($this->table->sort_order_column) && empty($this->{$this->table->sort_order_column})) {
			$this->{$this->table->sort_order_column} = $this->table->sort_order_default_pos;
		}
		
		// Add the values of the fields to be inserted to the SQL query
		$values_to_insert = array();
		if (count($fields_to_insert) > 0) {
			foreach ($fields_to_insert as $field) {
				if ($this->$field === NULL) {
					$value = 'NULL';
				} else {
					$value = $this->db->quote($this->$field);
				}
				$values_to_insert[] = $value;
			}
		} else {
			throw new Exception('DBRow::insert() called but no fields to insert found');
		}
		
		$SQL = 	'INSERT INTO '.$this->table->name.' ('.implode(',',$fields_to_insert).') VALUES ('.implode(',',$values_to_insert).')';
		
		### PHPOF-DYNAMIC-END: DBROW-INSERT-SQL ###
		
		$res = $this->db->query($SQL);
		if (PEAR::isError($res)) {
    		throw new Exception($res->getMessage());
		}
		
		// Update the ID if only a single key and the key hasnot been set manualy
		if (count($this->pri_keys) == 1 && $this->table->key_auto_increment && !isset($this->{$this->pri_keys[0]})) {
			$this->{$this->pri_keys[0]} = $this->db->lastInsertId($this->table->name);
		}
		
		if (!empty($this->table->sort_order_column)) $this->processSortOrder();
		$this->initChangeRecording();
	}
	
	/**
	 * Update row in database using the values currently stored in the properties of the object
	 *
	 * @return bool
	 */
	public function update ()
	{
		if (count($this->pri_keys) == 0) {
			throw new Exception('DBRow::update() called but no primary key for table');
			return false;
		}
		if ($this->primaryKeyFilled()) {
			// Handle LOB fields
			foreach ($this->fields_lob as $field) {
				if ($this->$field == PHPOF2_DB::LOB_DELETEPENDING) {
					call_user_func_array(array($this->driver, 'removeLOBData'), array(&$this, $field));
					$this->$field = null;
				}
			}
			
			// Create the SQL query
			$insert_values = array();
			$fields_to_update = $this->fields_normal;
			
			// If the "original" primary key isn't set, use the current one 
			if ($this->_orig_prikey === null) {
				$this->_orig_prikey = $this->serialisePriKey();
			} else {
				// If we know what the original pri key of this row was, check to
				// see if it has changed: if so, we need to update the pri key
				// columns too
				if ($this->_orig_prikey != $this->serialisePriKey()) {
					$fields_to_update = array_merge($fields_to_update, $this->pri_keys);
				}
			}
			
			foreach ($fields_to_update as $field) {
				if ($this->$field === NULL) {
					$value = 'NULL';
				} else {
					$value = $this->db->quote($this->$field);
				}
				$insert_values[] = $this->db->quoteIdentifier($field).'='.$value;
			}
			// TODO: only update changed cols for efficiency
			$SQL = 'UPDATE '.$this->table->name.' SET '.implode(',',$insert_values).' WHERE ' . $this->getPrikeyWhereClause($this->_orig_prikey);
			$res = $this->db->query($SQL);
			if (PEAR::isError($res)) {
	    		throw new Exception($res->getMessage());
			}	
		} else {
			if (count($this->pri_keys) > 0) {
				throw new Exception('PHPOF2_DBRow::Update() called, but the primary key is empty');
			} else {
				throw new Exception("PHPOF2_DBRow::Update() called, but table ($this->table->name) does not have a primary key field");
			}
		}
		
		if (!empty($this->table->sort_order_column)) $this->processSortOrder();
		$this->initChangeRecording();
		return true;
	}
	
	/**
	 * Delete current row from database
	 * 
	 * @return bool
	 */
	public function delete ()
	{
		if (count($this->pri_keys) > 0) {
			if ($this->primaryKeyFilled()) {
				
				// Remove any binary data from columns in the row
				if (count($this->fields_lob) > 0) {
					foreach ($this->fields_lob as $field) {
						call_user_func_array(array($this->driver, 'removeLOBData'), array(&$this, $field));
					}
				}
				
				// Remove the actual row from the database
				$res = $this->db->query('DELETE FROM ' . $this->table->name . ' WHERE '.$this->getPrikeyWhereClause());
				if (PEAR::isError($res)) {
	    			throw new Exception($res->getMessage());
				}
			} else {
				throw new Exception('PHPOF2_DBRow::delete() called but primary key is empty');
			}
		} else {
			// No table key - delete rows which match ALL (non-LOB) columns
			$sql = 'DELETE FROM ' . $this->table->name . ' WHERE ';
			foreach ($this->fields_normal as $field) {
				$clauses = array();
				if ($this->$field === NULL) {
					$clauses[] = $field . ' IS NULL';
				} else {
					$clauses[] = $this->db->quoteIdentifier($field) . '=' . $this->db->quote($this->$field);
				}
			}
			$sql .= implode(' AND ', $clauses);
			$res = $this->db->query($sql);
			if (PEAR::isError($res)) {
	    		throw new Exception($res->getMessage());
			}
		}
		$this->clear();

		return TRUE;
	}
	
	
	// usually called from PHPOF2_DBResult class
	public function fillFromResult(MDB2_Result &$result)
	{
		$row = $result->fetchRow(MDB2_FETCHMODE_ASSOC);
		if ($row === NULL) {
			// end of iteration
			return false;
		}
		if (!PEAR::isError($row)) {
			foreach ($row as $field => $value) {
				$this->$field = $value;
			}
			$this->resolveLOBFields();
			$this->initChangeRecording();
			return true;
		} else {
			throw new Exception($row->getMessage());
			return false;
		}
	}
	
	
	/**
	 * This function should really be overloaded by derived classes
	 * If not, we make a random guess at what might make a good description
	 */
	public function getDescriptiveName()
	{
		$possible_cols = array('name','title','description','desc');
		foreach($possible_cols as $col) {
			if (isset($this->$col)) {
				return $this->$col;
			}
		}
		
		throw new Exception('No descriptive field found for '.$this->table->name.' in '.__METHOD__);
	}
	
	/**
	 * Return a "simple" object containing the current values (not including LOB fields) of data fields in this object
	 */
	public function getSimpleObject()
	{
		require_once 'PHPOF2/DBRow/Simple.php';
		
		$simple = new PHPOF2_DBRow_Simple($this);
		$fields = array_merge($this->pri_keys, $this->fields_normal);
		foreach($fields as $field){
			$simple->_labels[$field]=$this->table->getLabel( $field );
			$simple->$field = $this->$field;
		}
		return $simple;
	}
	

	// ---------------------------------------------------------------------------------
	// SORT ORDER STUFF
	// ---------------------------------------------------------------------------------
	
	/**
	 * Assuming the current table has an arbitrary "sort order" column defined, allow it to be "moved" by rearranging the sort orders within the table
	 * 
	 * @param string $Direction  The direction to move the current row in; should be a PHPOF2_DBRow::MOVE_* constant
	 * @param int    $Num        How many steps in $Direction to move the current row
	 */
	public function move($Direction, $Num=1)
	{
		if (empty($this->table->sort_order_column)) {
			throw new Exception('DBRow::move() called but object\'s table does not have a sort column defined');
		}
		
		if ($Num > 1) {
			$this->move($Direction, ($Num-1));
		}
		
		// This is an additional qualifier for if a table has got several different sequences of sort
		// orders within it
		$Qualifier = $this->_getSortOrderQualifierSQL();
		
		switch ($Direction) {
			case self::MOVE_UP:
				$WhereClause = '<';
				$OrderDirection = 'DESC';
			break;
			
			case self::MOVE_DOWN:
				$WhereClause = '>';
				$OrderDirection = 'ASC';
			break;
			
			default:
			throw new Exception("Invalid direction in PHPOF2_DBRow::move() [Supplied: $Direction]");
		}
		
		// First see if the ordering is fully expanded (i.e. all items have a unique sort order value)
		$TotalItems = $this->db->queryOne("SELECT COUNT(*) AS count FROM ".$this->table->name." WHERE 1 $Qualifier");
		
		$UniqueItems = $this->db->queryOne("SELECT COUNT(DISTINCT ".$this->table->sort_order_column.") AS count FROM ".$this->table->name." WHERE 1 $Qualifier");
		
		if ($TotalItems != $UniqueItems) {
			if (empty($this->table->_key)) {
				throw new Exception('DBRow::sort() called, but the table (with qualifiers) does not contain a set of unique sort values and the table does not have a primary key column. The behaviour of this function is undefined.');
			}
			$this->_sort($Qualifier);
		}
		$AdjacentItem = $this->db->query("SELECT " . implode(', ',$this->pri_keys) . ", $this->fieldlist FROM ".$this->table->name." WHERE ".$this->table->sort_order_column." $WhereClause ".$this->{$this->table->sort_order_column}." $Qualifier ORDER BY ".$this->table->sort_order_column." $OrderDirection LIMIT 1");
		if (PEAR::isError($AdjacentItem)) {
    		throw new Exception($AdjacentItem->getMessage());
		}
		if ($AdjacentItem->numRows() > 0) {
			$Item = new PHPOF2_DBRow($this->table);
			
			// Move the adjacent item to the current item's position
			$AdjacentItem->fetchIntoObject($Item);
			$AdjacentOrder = $Item->{$this->table->sort_order_column};
			$Item->{$this->table->sort_order_column} = $this->{$this->table->sort_order_column};
			$Item->update();
			
			// Move current item to its final position
			$this->{$this->table->sort_order_column} = $AdjacentOrder;
			$this->update();
		} else {
			// Already at the top
			return false;
		}
	}
	
	protected function _getSortOrderQualifierSQL()
	{
		$Qualifier = '';
		if (!empty($this->table->sort_order_qualifier_columns) && is_array($this->table->sort_order_qualifier_columns)) {
			foreach ($this->table->sort_order_qualifier_columns as $ColumnName) {
				$Qualifier = "AND $ColumnName=".$this->$ColumnName;
			}
		}
		return $Qualifier;
	}
	
	protected function processSortOrder()
	{
		$SortField = $this->table->sort_order_column;
		if (!empty($this->$SortField) && !empty($this->table->_key)) {
			if ($this->$SortField == DBROW_SORT_TOP || $this->$SortField == DBROW_SORT_BOTTOM) {
				$Qualifier = $this->_getSortOrderQualifierSQL();
				if ($this->$SortField == DBROW_SORT_TOP) {
					// Sort to top by setting a priority ordering factor of "<key>=<id> DESC"
					$this->_sort($Qualifier, $this->table->_key.'='.$this->id.' DESC');
				} else {
					// Sort to bottom
					$Res = $this->db->query("SELECT MAX($SortField) AS max FROM ".$this->table->name." WHERE 1 $Qualifier");
					if (PEAR::isError($Res)) {
			    		throw new Exception($Res->getMessage());
					}
					$Row = $Res->fetchRow();
					$this->$SortField = $Row['max']+1;
					$this->update();
				}
			}
		}
	}
	
	protected function _sort($Qualifier, $PriorityOrder='')
	{
		if (!empty($PriorityOrder)) $PriorityOrder.=',';
		$AllItems = $this->db->query('SELECT * FROM '.$this->table->name." WHERE 1 $Qualifier ORDER BY $PriorityOrder ".$this->table->sort_order_column.' ASC');
		if (PEAR::isError($AllItems)) {
    		throw new Exception($AllItems->getMessage());
		}
		
		if ($AllItems->numRows() > 0) {
			$Item = new PHPOF2_DBRow($this->table);
			$OrderValue = 1;
			while ($AllItems->fetchIntoObject($Item)) {
				if ($Item->{$this->table->_key} == $this->{$this->table->_key}) {
					$NewOrderValue = $OrderValue;
				}
				$Item->{$this->table->sort_order_column} = $OrderValue++;
				$Item->update();
			}
			
			// Update this object with it's new sort order value
			$this->{$this->table->sort_order_column} = $NewOrderValue;
		}
	}
	
	// ---------------------------------------------------------------------------------
	// CHANGE RECORDING
	// ---------------------------------------------------------------------------------
	
	/**
	 * Set the initial values for change recording
	 * Does not handle LOB fields
	 *
	 * @return void
	 */
	public function initChangeRecording()
	{
		// This is nothing to do with change recording but is placed here
		// for convenience and thus is called unconditionally
		$this->_orig_prikey = $this->serialisePriKey();
		
		if (!$this->record_changes) return;
		// Clear the current set of data
		$this->original_values  = array();
		
		// Store the current data
		foreach ($this->fields_normal as $field) {
			$this->original_values[$field] = $this->{$field};
		}
	}
	
	/**
	 * Retrieve a list of fields which have changed since the object was 
	 * instantiated. Does not handle LOB fields.
	 *
	 * @return array Array of changed fields, in the format "fieldname" => "newvalue"
	 */
	public function getChanges()
	{
		if (is_array($this->original_values)) {
			$changes = array();
			foreach ($this->fields_normal as $field) {
				if ($this->$field != $this->original_values[$field]) {
					$changes[$field] = $this->$field;
				}
			}
			foreach ($this->fields_lob as $field) {
				if ($this->$field == PHPOF2_DB::LOB_DELETEPENDING) {
					$changes[$field] = null;
				}
			}
			return $changes;
		} else {
			//TODO Should load a copy of the object if the key is set and then do the comparison
			throw new Exception('Cannot get changes - the original values were not stored');
		}	
	}
}

