<?php

/**
 * PHPOF2/DBTable.class.php - PHPOF2 database table class - Table Data Gateway
 *
 * Copyright 2001-2010 Tim Jackson (tim@timj.co.uk)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU Lesser General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @package PHPOF2
 *
 */

class PHPOF2_DBTable {

	const      SORT_TOP = 'top';
	const      SORT_BOTTOM = 'bottom';

	const      USE_DN = 'usedn';
	const      USE_KEY = 'usekey';

	const      ORDER_DEFAULT = '!deforder!';

	const      COMPKEY_SEPARATOR = '#!#'; // The separator used when serializing composite keys

	const      DEFAULT_ROW_CLASS_AUTO = 'auto';

	public     $name;	// Table name in database

	// this is public, but use by random other classes (other than DBRows which are part of this table)
	// without a particular reason is discouraged
	public     $db; // access to underlying DB object
	
	/**
	 * @var bool Whether the primary key is auto-incrementing or not
	 */
	public     $key_auto_increment = false;

	// Whether to add new rows to top or bottom of table.
	// Set to PHPOF2_DBTable::SORT_BOTTOM or PHPOF2_DBTable::SORT_TOP.
	public     $sort_order_default_pos = PHPOF2_DBTable::SORT_BOTTOM;

	// The column containing an integer value which represents the (arbitrary) sort order of the rows
	// TODO: implement use of the default_order property throughout including in DB_result possibly?
	public     $sort_order_column = ''; //used by dbrow
	
	// If you intend to use DBRow::move() and you need the re-ordering to be 
	// restricted to a particular group of rows which are identified by a
	// specific column value (e.g. a category ID or something), list the names
	// of the columns that are used to constrain the move in this array
	public     $sort_order_qualifier_columns = array();

	/**
	 * @var array Array containing names of all columns except LOB fields, in the format: 
	 *            array('field1','field2')
	 *            The list of fields INCLUDES keys
	 */
	protected  $fields_normal = array();
	
	/**
	 * @var array Array containing names of all LOB fields, in the format: 
	 *            array('field1','field2')
	 */
	protected  $fields_lob = array();
		
	/**
	 * @var array List of all primary key fields, in the format:
	 *            array('field1','field2')
	 */
	protected  $pri_keys = array();
	
	/**
	 * @var array Details of all foreign key constraints including details of what field 
	 *            the key references and which table the referenced field is in,
	 *            in the format:
	 *
	 *            array(colname => array (referenced_table.col [, referenced_table.col [, ...] ])
	 */
	protected  $fgn_keys = array();
	
	/**
	 * Cached copies of some of the above field lists, all comma-separated and
	 * DB-escaped
	 */
	protected  $pri_key_list = ''; // Corresponds to $pri_keys
	protected  $fieldlist;         // Corresponds to $fields_normal

	/**
	 * @var string The default class to use when createRowObject() is called.
	 *             If set to self::DEFAULT_ROW_CLASS_AUTO, then if table 
	 *             auditing is enbled (@see $audit_logging) it will be 
	 *             'PHPOF2_DBRow_Audited', otherwise 'PHPOF2_DBRow'
	 */
	protected  $default_row_class = self::DEFAULT_ROW_CLASS_AUTO;
	protected  $default_order = '';       // how to sort the table, by default
	
	private    $field_defs = array();     // MDB2 field definitions
	
	protected  $table_label = null;	//for overriding the user-presented table name
	protected  $table_singular_label = null;		//for overriding the user-presented table-singular (row) name
	protected  $column_descriptions = array();
	
	/**
	 * @var bool Whether to enable audit logging
	 */
	protected  $audit_logging = false;

	/**
	 * @var array List of columns to ignore for auditing purposes, if audit
	 *            logging is enabled (@see $audit_logging)
	 */
	protected  $audit_columns_ignore = array();
	
	/**
	 * @var string The type of LOB storage to use by default for this table. Should be a PHPOF2_DB::LOB_STORAGE_* constant. If set to NULL, will inherit from the option 'lob_storage' set in the PHPOF2_DB object
	 */
	protected  $lob_storage = null;

	/**
	 * @var array Any per-column overrides to the default LOB storage preference for this table, in the format array('colname' => self::LOB_STORAGE_*)
	 */
	protected  $lob_storage_overrides = array();
	
	/**
	 * @var string If you are using LOB_STORAGE_FS for any columns, this specifies where on the FS the LOBs will be stored. Can optionally contain a %t for the table name and/or a %c for the column name. If set to NULL, will inherit from the DB option 'lob_storage_fs_path' if set
	 */
	protected  $lob_storage_fs_path = null;
	
	/**
	 * Constructor
	 */
	public function __construct (MDB2_Driver_Common &$db, $name)
	{
		if (!$name) {
			throw new Exception('PHPOF2_DBTable constructor called with an empty table name');
			return false;
		}
		$this->db = &$db;
		$this->name = $name;

		// Find out what this table looks like
		$this->discoverSchema();

		// Set how LOBs are stored
		if ($this->lob_storage === null) {
			$this->lob_storage = $this->db->phpof_options['lob_storage'];
		}

		return true;
	}
	
	
	/**
	 * Factory method to create a row object for this table
	 * NB: protected  $default_row_class needs to be overriden if not PHPOF2_Row
	 * 
	 * @param int $key_value=null
	 * @return row object
	 */
	public function createRowObject($key_value = NULL)
	{
		$row_object = $this->default_row_class;
		if ($row_object == self::DEFAULT_ROW_CLASS_AUTO) {
			if ($this->audit_logging) {
				require_once 'PHPOF2/DBRow/Audited.php';
				$row_object = 'PHPOF2_DBRow_Audited';
			} else {
				$row_object = 'PHPOF2_DBRow';
			}
		}
		// TODO? Require() file with object?
		return new $row_object($this, $key_value);
	}

	/**
	 * Return an array containing the primary keys in the table
	 */
	public function getPrimaryKeys()
	{
		return $this->pri_keys;
	}

	/**
	 * Returns an array containing all the foreign keys in the table (along with details of the table/key it references)
	 */
	public function getForeignKeys()
	{
		return $this->fgn_keys;
	}

	/**
	 * Return array containing list of column names of non-LOB and non-key columns in the table
	 */
	public function getFieldListNormal()
	{
		return $this->fields_normal;
	}
	
	/**
	 * return the full field list inc PKs, but barring LOBs.
	 *
	 * @return array $this->fieldlist
	 */
	public function getFieldListPKs(){
		return explode(',', str_replace('`', '', $this->fieldlist));
	}
	
	/*
	* Return an array containing a list of the column names of the LOB columns
	* in the table
	*/
	public function getFieldListLOB()
	{
		return $this->fields_lob;
	}
	
	/**
	 * Return how the specified LOB column stores its data
	 *
	 * @param string $data_column  Data column of interest
	 * @return string A PHPOF2_DB::LOB_STORAGE_* const
	 * @throws Exception if $data_column is not a LOB column 
	 */
	public function getLOBStorageType($data_column)
	{
		if (!in_array($data_column, $this->fields_lob)) {
			throw new Exception("Column $data_column is not a LOB field");
		}
		if (isset($this->lob_storage_overrides[$data_column])) {
			return $this->lob_storage_overrides[$data_column];
		}
		return $this->lob_storage;
	}
	
	/**
	 * Returns the path in which LOB data files will be stored, for data stored on the filesystem
	 *
	 * @param string $col    The column for which data is being stored
	 * @return string|false  False if LOB storage is not enabled for this table
	 */
	public function getLOBStorageFSPath($col)
	{
		if ($this->lob_storage_fs_path) {
			return $this->lob_storage_fs_path;
		} else {
			if (isset($this->db->phpof_options['lob_storage_fs_path'])) {
				$path = str_replace('%t', $this->name, $this->db->phpof_options['lob_storage_fs_path']);
				$path = str_replace('%c', $col, $path);
				return $path;
			}
		}
		return false;
	}
	
	/**
	 * Migrate the storage of LOBs in the specific column(s) in this table to a different LOB storage type
	 *
	 * @param string $storage_type  A PHPOF2_DB::LOB_STORAGE_* constant
	 * @param array|null $columns   List of columns to migrate. If NULL, all columns.
	 */
	public function migrateLOBStorage($storage_type, $columns=null)
	{
		if ($columns === null) $columns = $this->fields_lob;
		$all = $this->getAll();
		$row = $this->createRowObject();
		while ($all->fetchIntoObject($row)) {
			foreach ($columns as $column)
			{
				if (!in_array($column, $this->fields_lob)) {
					throw new Exception("Error when migrating LOB storage: $column is not a LOB column");
				}
				$data = $row->getLOBData($column);
				$row->storeLOBData($data, $column, $storage_type);
			}
		}
	}
	
	/**
	 * Return the total number of rows in the table
	 *
	 * @return int
	 */
	public function numRows()
	{
		$res = $this->db->query("SELECT COUNT(*) AS c FROM $this->name");
		$row = $res->fetchRow();
		return $row[0];
	}
	
	/**
	 * Return whether or not the table is being audited
	 * @return bool
	 */
	public function auditTable()
	{
		return $this->audit_logging;
	}

	/**
	 * Return an array of fields in this table which will be audited
	 * This just gives a sensible default; it makes sense to overload this in derived classes.
	 */
	public function getAuditableFields()
	{
		// Remove primary keys from fields_normal
		$audit = array_diff($this->fields_normal, $this->pri_keys);
		
		// Check if the auditable fields exist
		if (is_array($this->audit_columns_ignore)) {
			$audit = array_diff($this->fields_normal, $this->audit_columns_ignore);
		}	
		return $audit; 
	}

	/**
	 * Discover the structure of the table and store it internally in the
	 * following properties:
	 *  - fields_lob
	 *  - fields_normal
	 *  - pri_keys
	 *  - fgn_keys
	 *  - key_auto_increment
	 * 
	 * Note that foreign key discovery only works with MDB2 > 2.5.0b1 (unless
	 * the optimised driver bypasses MDB2)
	 * 
	 * If you have a table where the schema changes infrequently and you need
	 * improved performance whilst still gaining the benefits of PHPOF2, you may
	 * wish to overload this function and statically fill the above properties.
	 * If you do that, you will probably want to still call 
	 * self::_createSchemaCachedLists() to create the *_list properties for you.
	 *
	 * @return void
	 */
	private function discoverSchema()
	{
		// Run optimisation, if it exists
		$optimisation_class = 'PHPOF2_Optimisation_' . $this->db->phptype;
		if (class_exists($optimisation_class)) {
			$ret = call_user_func_array(array($optimisation_class, 'tableDiscoverSchema'), array(
				&$this,
				&$this->fields_normal,
				&$this->fields_lob,
				&$this->pri_keys,
				&$this->fgn_keys,
				&$this->key_auto_increment)
			);
			if ($ret != PHPOF2_Optimisation_Interface::UNIMPLEMENTED) {
				return;
			}
		}
		// we need these modules from MDB2
		$this->db->loadModule('Manager');
		$this->db->loadModule('Reverse');
		
		$fields = $this->db->manager->listTableFields($this->name);
		if(PEAR::isError($fields)){
			throw new Exception("Error getting columns for ".$this->name);
		}
		$autoinc = array();
		foreach ($fields as $field) {
			$field_defs = $this->db->reverse->getTableFieldDefinition($this->name, $field);

			if (PEAR::isError($field_defs)) {
				throw new Exception($field_defs->getMessage()."(when getting definition for ".$this->name.".$field)");
			}
			// a DB type may map to more than one MDB2 type, in which case
			// there may be multiple entries in $field_defs.
			// We ignore this and just use the default mapping (index 0)
			if ($field_defs[0]['type'] == 'blob' || $field_defs[0]['type'] == 'clob') {
				$this->fields_lob[] = $field;
			} else {
				$this->fields_normal[$field] = $field; // we use $field as the index as well as value so we can remove elements easily later
			}
			if (isset($field_defs[0]['autoincrement']) && $field_defs[0]['autoincrement'] === true) {
				$autoinc[$field] = true;
			}
			$this->field_defs[$field] = $field_defs;
		}

		$constraints = $this->db->manager->listTableConstraints($this->name);
		foreach ($constraints as $constraint_name) {	//these include indexes, foreign keys etc.
			// MDB2_Driver_mysqli 2.5.0b1 is completely and utterly broken - see PEAR bug #13412
			$constraint_details = $this->db->reverse->getTableConstraintDefinition( $this->name, $constraint_name );
			if (PEAR::isError($constraint_details)) {
				throw new Exception ("MDB2 error whilst getting '$constraint_name' constraint for table '$this->name': ".$constraint_details->getMessage());
			}
			if (isset($constraint_details['primary']) && $constraint_details['primary'] == true) {
				// contains e.g. "'fields' => array('keyname' => array('sorting' => 'ascending'))";
				$this->pri_keys = array_keys($constraint_details['fields']);
			} else if (isset($constraint_details['foreign']) && $constraint_details['foreign'] == true) {
				/*
				 * See MDB2/Driver/Reverse/mysqli.php to see the structure - it's:
				   array('foreign' => true,
				         'fields' => array(
				             localcol1 => array('position' => pos),
				             localcol2 => array('position' => pos)
				         )
                         'references' => array(
				             'table' => referenced_table_name,
					         'fields' => array(
				                refcol1 => array('position' => pos),
				                refcol2 => array('position' => pos)
				             )
	                      )
	                 )
				 */
				// Turn the referenced column array around to get the positions
				$refcol_index = array();
				foreach ($constraint_details['references']['fields'] as $refcol => $refcol_details) {
					$refcol_index[$refcol_details['position']] = $refcol;
				}
				// this assumes that MDB2 puts the things in the 'fields' array
				// in the same order as in 'references'['fields'] (which it does)
				foreach ($constraint_details['fields'] as $field => $field_details) {
					$this->fgn_keys[$field][] = $constraint_details['references']['table'] . '.' . $refcol_index[$field_details['position']];
				}
			}
		}
		
		foreach ($this->pri_keys as $key) {
			// remove from the list of "normal" fields
			unset($this->fields_normal[$key]);
		}
		$this->fields_normal = array_values($this->fields_normal);	//remove duplicates: the short-hand used for field removal (the keys)
		
		// If there is a single primary key and it's auto-incrementing, flag this
		if (count($this->pri_keys) == 1 && isset($autoinc[$this->pri_keys[0]])) {
			$this->key_auto_increment = true;
		}
		
		// Now create _list properties
		$this->_createSchemaCachedLists();
	}
	
	/**
	 * Create cached properties containing lists of *pre-quoted* fields and
	 * primary keys. These are useful for speed when generating
	 * SQL queries
	 * This is public so it can be used by optimisation functions
	 * 
	 * @return void
	 */
	public function _createSchemaCachedLists()
	{
		$fieldlist = array();
		foreach (array_merge($this->pri_keys, $this->fields_normal) as $field) {
			$fieldlist[] = $this->db->quoteIdentifier($field);
		}
		$this->fieldlist = implode(',', $fieldlist);

		$fieldlist = array();
		foreach ($this->pri_keys as $field)	{
			$fieldlist[] = $this->db->quoteIdentifier($field);
		}
		$this->pri_key_list = implode(',',$fieldlist);
		
		return;
	}

	/**
	 * Get a string-serialised version of the NAMES of the primary key(s) of
	 * the table. This works for composite keys
	 */
	public function serialisePriKey()
	{
		if (count($this->pri_keys) == 1) {
			return $this->pri_keys[0];
		} else {
			return implode(self::COMPKEY_SEPARATOR, $this->pri_key);
		}
	}

	/**
	 * Return an associative array containing NAME=>VALUE of the primary keys of
	 * the table.
	 *
	 * @param string $value Serialised primary key value
	 */
	public function unserialisePriKey($value)
	{
		$values = explode(self::COMPKEY_SEPARATOR, $value);
		foreach ($this->pri_keys as $index => $key_col)
		{
			if (!isset($values[$index])) {
				throw new Exception('unserialise failed - serialised key does not have correct number of key fields');
			}
			$output[$key_col] = $values[$index];
		}
		return $output;
	}


	// ------------------------------------------------------
	// Array generation functions

	/**
	 * Create an array containing matching $key => $value pairs from the table
	 *
	 * @param string $key     Column to use as the array index, or PHPOF2_DBTable::USE_KEY, in which case the table's (serialised) primary key will be used
	 * @param string $value   Column to use as the array value, or PHPOF2_DBTable::USE_DN in which case each row is fetched and getDescriptiveName() is called on it
	 * @param string $orderby  A fragment to be used in an SQL 'ORDER BY' clause, or PHPOF2_DBTable::ORDER_DEFAULT in which case the table's default ordering is used
	 */
	public function getArray ($key=PHPOF2_DBTable::USE_KEY, $value=PHPOF2_DBTable::USE_DN, $orderby=PHPOF2_DBTable::ORDER_DEFAULT)
	{
		if ($key == PHPOF2_DBTable::USE_KEY) {
			if (count($this->pri_keys) == 0) {
				throw new Exception('DBTable::getArray() called but table has no primary key');
			}
			$key = $this->serialisePriKey(); // the serialised identifier to use in output array
			$key_to_select = $this->pri_key_list;// what rows to actually select from DB
		} else {
			$key_to_select = $key;
		}

		// If 'value' is set to PHPOF2_DBTable::USE_DN, get the descriptive name
		if ($value == PHPOF2_DBTable::USE_DN) {
			$row =& $this->createRowObject();
			$AllRows = $this->getAll($orderby);
			if ($AllRows->numRows() > 0) {
				while ($AllRows->fetchIntoObject($row)) {
					$ReturnArray[$row->{$key}] = $row->getDescriptiveName();
				}
				return $ReturnArray;
			}
			return array();
		} else {
			if ($orderby == PHPOF2_DBTable::ORDER_DEFAULT) {
				$orderby = $this->getOrderPreference();
			}
			$res = $this->db->query("SELECT $key_to_select, $value FROM $this->name ORDER BY $orderby");
			if (PEAR::isError($res)) {
	    		throw new Exception($res->getMessage());
			}
			return $res->getArray($key, $value);
		}
	}

	function getArray_Paged ($key, $value, $results_per_page, $page, $orderby=PHPOF2_DBTable::ORDER_DEFAULT)
	{
		if ($orderby == PHPOF2_DBTable::ORDER_DEFAULT) $orderby = $this->getOrderPreference($value);
		$this->db->setLimit($results_per_page, (($page -1)  * $results_per_page));
		if ($this->db->supports('limit_queries') === 'emulated') {
			trigger_error("MDB2 driver does not fully support paged queries", E_USER_NOTICE);
		}
		$res = $this->db->query("SELECT $key, $value FROM $this->name ORDER BY $orderby");
		return $res->getArray($key, $value);
	}

	/**
	 * Find the preferred ordering for this table in a format to be used in an 
	 * SQL 'ORDER BY' clause, for example "col1" or "col1 ASC"
	 * The preference is found by taking the following possible sources
	 * of ordering information in turn until a non-empty one is found
	 * - the default_order property
	 * - the sort_order_column column (ascending)
	 * - the supplied fallback (if any)
	 * - the integer 1
	 * 
	 * @param string $Fallback  A fallback column to use if no other sort order can be ascertained
	 */
	public function getOrderPreference($fallback=null)
	{
		if (empty($this->default_order)) {
			if (!empty($this->sort_order_column)) {
				return $this->sort_order_column . ' ASC';
			} else if (!empty($fallback)) {
				return $fallback;
			} else {
				return '1';
			}
		}
		return $this->default_order;
	}


	// ------------------------------------------------------
	// Search functions

	/**
	 * Return all rows in the table
	 *
	 * @param string $orderby  An SQL 'ORDER BY' clause
	 * @return PHPOF2_DBResult_Interface
	 */
	public function getAll ($orderby=PHPOF2_DBTable::ORDER_DEFAULT)
	{
		return $this->_getAll ($orderby);
	}

	/**
	 * Return all rows in the table, split into pages
	 *
	 * @param int $results_per_page  Number of "pages" in the result set
	 * @param int $page              The page offset
	 * @param string $orderby        An SQL 'ORDER BY' clause
	 * @return PHPOF2_DBResult_Interface
	 */
	function getAll_paged ($results_per_page, $page=1, $orderby=PHPOF2_DBTable::ORDER_DEFAULT)
	{
		if ($page == 0) $page = 1;
		return $this->_getAll ($orderby, $results_per_page, $page);
	}

	protected function _getAll ($orderby = PHPOF2_DBTable::ORDER_DEFAULT, $results_per_page = 0, $page = 1)
	{
		if ($orderby == PHPOF2_DBTable::ORDER_DEFAULT) {
			$orderby = $this->getOrderPreference();
		}
		$sql = "SELECT $this->fieldlist";
		if (count($this->pri_keys) > 0) $sql .= ','.implode(',',$this->pri_keys);
		$sql .= ' FROM '.$this->name;
		if ($orderby) $sql .= " ORDER BY $orderby";
		if ($results_per_page > 0) {
			$this->db->setLimit($results_per_page, ($page-1)*$results_per_page);
		}
		$res = $this->db->query($sql);
		if (PEAR::isError($res)) {
    		throw new Exception($res->getMessage());
		}
		return $res;
	}

	/**
	 * Do a basic table search, looking for $field == $value
	 * If passed an array in $value, it performs an "IN" search
	 * 
	 * @param string $field The name of the column to search in
	 * @param string|array $value The value(s) to search for
	 * @param string $orderby The order clause to use. If not set, uses table default
	 * @return PHPOF2_DBResult_Interface
	 */ 
	public function basicSearch ($field, $value, $orderby = PHPOF2_DBTable::ORDER_DEFAULT)
	{
		return $this->_search('=', $field, $value, $orderby);
	}
	
	/**
	 * Do a basic table search, looking for $field == $value
	 * If passed an array in $value, it performs an "IN" search
	 * 
	 * @param string $field          The name of the column to search in
	 * @param string|array $value    The value(s) to search for
	 * @param int $results_per_page  Number of "pages" in the result set
	 * @param int $page              The page offset
	 * @param string $orderby        The order clause to use. If not set, uses table default
	 * @return PHPOF2_DBResult_Interface
	 */ 
	function basicSearch_Paged ($field, $value, $results_per_page, $page, $orderby = PHPOF2_DBTable::ORDER_DEFAULT)
	{
		return $this->_search('=', $field, $value, $orderby, $results_per_page, $page);
	}

	/**
	 * Do a basic table search, looking for $field != $value
	 * If passed an array in $value, it performs a "NOT IN" search
	 * 
	 * @param string $field The name of the column to search in
	 * @param string|array $value The value(s) to search for
	 * @param string $orderby The order clause to use. If not set, uses table default
	 * @return PHPOF2_DBResult_Interface
	 */ 
	function basicSearchNot ($field, $value, $orderby = PHPOF2_DBTable::ORDER_DEFAULT)
	{
		return $this->_search('!=', $field, $value, $orderby);
	}
	function basicSearchNot_Paged ($field, $value, $results_per_page, $page, $orderby = PHPOF2_DBTable::ORDER_DEFAULT)
	{
		return $this->_search('!=', $field, $value, $orderby, $results_per_page, $page);
	}

	/**
	 * Do a basic table search, looking for $field LIKE $value
	 * 
	 * @param string $field The name of the column to search in
	 * @param string|array $value The value(s) to search for
	 * @param string $orderby The order clause to use. If not set, uses table default
	 * @return PHPOF2_DBResult_Interface
	 */ 
	public function basicSearchLike ($field, $value, $orderby = PHPOF2_DBTable::ORDER_DEFAULT) {
		return $this->_search('LIKE', $field, $value, $orderby);
	}
	function basicSearchLike_Paged ($field, $value, $results_per_page, $page, $orderby = PHPOF2_DBTable::ORDER_DEFAULT) {
		return $this->_search('LIKE', $field, $value, $orderby, $results_per_page, $page);
	}

	/**
	 * The field-value pairs are used to build a like for like comparison query,
	 * the result set of which is returned.
	 * Optionally, the fields to be returned can be specified and/or the results
	 * ordered.
	 * 
	 * @param array $field_value_pairs  An associative array of columns to search for, with their corresponding values
	 * @param array $fields_to_select   Which columns to return in the result. If NULL, all non-binary columns are returned
	 * @param string $orderby           A fragment to be used in an SQL 'ORDER BY' clause, or PHPOF2_DBTable::ORDER_DEFAULT in which case the table's default ordering is used
	 * @return PHPOF2_DBResult_Interface
	 */
	public function searchValues(array $field_value_pairs, $fields_to_select = null, $orderby = PHPOF2_DBTable::ORDER_DEFAULT)
	{
		if ($orderby == PHPOF2_DBTable::ORDER_DEFAULT) {
			$orderby = $this->getOrderPreference();
		}
		// Check the input
		if (count($field_value_pairs) == 0) {
			throw new Exception('Field/value pairs array passed to PHPOF2_DBTable::find() was empty');
		}

		// Construct the WHERE clause
		$conditions = array();
		foreach ($field_value_pairs as $field_tmp => $value_tmp){
			$conditions[] = $this->db->quoteIdentifier($field_tmp) . '=' . $this->db->quote($value_tmp);
		}
		$where_clause = implode(' AND ', $conditions);

		// Determine which fields to select
		if (is_array($fields_to_select) && count($fields_to_select) > 0) {
			$fields_to_select_array = array();
			foreach ($fields_to_select as $field) {
				$fields_to_select_array[] = $this->db->quoteIdentifier($field);
			}
			$fields_to_select = implode(',', $fields_to_select_array);
		} else {
			$fields_to_select = $this->fieldlist;
		}

		$sql = "SELECT $fields_to_select FROM $this->name WHERE $where_clause";

		if ($orderby) $sql .= " ORDER BY $orderby";

		return $this->db->query($sql);
	}

	protected function _search ($clause, $field, $value, $orderby = PHPOF2_DBTable::ORDER_DEFAULT, $results_per_page = 0, $page = 0)
	{
		if ($orderby == PHPOF2_DBTable::ORDER_DEFAULT) {
			$orderby = $this->getOrderPreference();
		}
		
		if (is_array($value)) {
			if ($clause == '=') {
				$clause = 'IN';
			} else {
				$clause = 'NOT IN';
			}
		}
		
		$sql = 'SELECT '.$this->fieldlist;
		$sql .= ' FROM '.$this->name." WHERE ".$this->db->quoteIdentifier($field) . " $clause ";

		if ($clause == 'IN' || $clause == 'NOT IN') {
			$val_array = array();
			foreach ($value as $val) {
				$val_array[] = $this->db->quote($val);
			}
			$sql .= '(' . implode(', ', $val_array) . ')';
		} else {
			$sql .= $this->db->quote($value);
		}
		if ($orderby) $sql .= " ORDER BY $orderby";
		
		if ($results_per_page > 0) {
			$this->db->setLimit($results_per_page, (($page -1)  * $results_per_page));
			if ($this->db->supports('limit_queries') === 'emulated') {
				trigger_error("MDB2 driver does not fully support paged queries", E_USER_NOTICE);
			}
		}
		$res = $this->db->query($sql);
		if (PEAR::isError($res)) {
    		throw new Exception($res->getMessage());
		}
		return $res;
	}
	
	
	/**
	 * Get a label for a column. Deprecated.
	 *
	 * @deprecated
	 * @see PHPOF2_DBRow::getLabel()
	 * @param string $col_name Column name
	 * @return string Column label
	 */
	public function getColumnDescription($field_name)
	{
		return $this->getLabel($field_name);
	}
	
	/**
	 * Get label for a given column
	 *
	 * @param string $field_name  Column name
	 * @return string  Column label
	 */
	public function getLabel($field_name){
		if (isset($this->column_descriptions[$field_name])) {
			$description = $this->column_descriptions[$field_name];
		} else {
			$description = $field_name;
			$description = preg_replace('/_id$/i','_ID',$description);
			$description = preg_replace('/_const$/i','',$description);
			$description = str_replace('_',' ',$description);
			$description = ucwords($description);
		}
		return $description;
	}
	
	/**
	 * Get text label for the current table
	 *
	 * @return string  Table label
	 */
	public function getTableLabel() {
		return ($this->table_label)? $this->table_label : str_replace('_',' ', ucwords($this->name));
	}
	
	public function getTableSingularLabel() {
		if ($this->table_singular_label) {
			return	$this->table_singular_label;
		} else {
			return preg_replace('/s(\s|$)/', '', preg_replace('/ies(\s|$)/', 'ys', $this->getTableLabel()));
		}
	}
	
}
